; Global constants

P_FLAG	EQU 23697
FLAGS2	EQU 23681
ATTR_P	EQU 23693	; permanet ATTRIBUTES
ATTR_T	EQU 23695	; temporary ATTRIBUTES
CHARS	EQU 23606 ; Pointer to ROM/RAM Charset
UDG	EQU 23675 ; Pointer to UDG Charset
MEM0	EQU 5C92h ; Temporary memory buffer used by ROM chars

