#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from .arrayaccess import SymbolARRAYACCESS


class SymbolARRAYLOAD(SymbolARRAYACCESS):
    """ This class is the same as SymbolARRAYACCESS, we just declare it to make
    a distinction. (e.g. the Token is gotten from the class name)
    """
    pass
