#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os

# Initialization
path = os.path.realpath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))
sys.path.insert(0, path)
